!+
! Module: ommodmap_mod
!
! Authors:  Vladimir Yershov - MSSL
! Authors:  Jon Rainger - MSSL
! Authors:  Phil Williams - MSSL
! Authors:  Matt Page - MSSL
!          
! Description: Matt's algorithm for dealing with fixed pattern noise
! 
! Comments: 
! 
! $Id: module.f90,v 1.2 2002/10/16 13:45:39 gvacanti Exp $
!- 
!
!- Modified 13th June 2000 - Corrected error messaging
!- Modified 4th June 2000 - Added section to tidy up memory
!- Modified 5th September 2000 - Ammended code to process windows that have size not modulo 8
!
!- 2002-10-16 (VNY) New parameter (flatSet) has been added in order to extract the Mod8 tile from
!             this file
!  2002-1--21 (VNY) New parameter (multiplyFlat) has been added in order to make it possible a different
!             method of mod-8 noise correction
! 2003-07-10 (VNY) check if the image is corrupted (possible missing columns)
! 2003-08-28 (VNY) the parameters applyFlat ad multiplyFlat are nor mandatory anymore
! 2003-09-29 (VNY) tested in the imaging mode with the visual and UV filters; the code is 
!                  cleaned, some small bugs are removed
! 2003-10-02 (VNY) It is found that the task is probably cycling infinitely on an image
!                  vith very few counts. The problem is caused by an unrealistically large
!                  values of two pixels (65535)
! 2003-10-07 (VNY) There was a bug found that causes infinite cycling of the task 
!                  (the allocation of the bad pixel map from the quality array was incorrect)
!                  (SSC-SPR-3156)
! 2003-10-16 (VNY) Small bugs are fixed 
! 2003-10-24 (VNY) A new boolean parameter (mod8correction) is introduced
! 2003-10-29 (VNY) A bug, resulting in the output image divided by the factor two, was fixed (SSC-SPR-3160)
! 2003-10-31 (VNY) Parameter mod8correction change its type from bool to integer
! 2003-11-26 (VNY) Check of the image-size consistency with the parameters
!                  DETSEC, windowx0, windowy0 is introduced 
! 2004-05-20 (VNY) The limiting pixel value was increased from 50000 to 250000
!                  (SSC-SPR-3323)
! 2004-06-09 (VNY) The problem od subscript in mod8TileDX and mod8TileDY mask arrays is fixed
!                  (the meaning of the variables xmax and ymax has been changed since long ago, 
!                  but in a piece of code they still were understood as dx and dy) 
! 2004-11-10 (VNY) Tests with the module recover16bits_mod
! 2005-02-23 (VNY) The value of MJDREF is set to 50814, in accordance with the changes 
!                  made 04.06.2004 in OMFASTSHIFT, SSC-SPR-3333. The old value (5.08140007312960D+04)
!                  was only used in rare occasions when the MJDREF key was abscent in the header.
!                  So, this change practically does not affect the performance of the task.
! 2007-07-23 (VNY) Protected against cycling when the image contains black strips: analysis of the 
!                  quality image is introduced and the modulo-8 mask is build only for good 
!                  pixels
! 2008-04-04 (VNY) Checking for the size of the bad-pixel region and, if this region is small 
!                  (currently checking the 60 by 60-pixel vertical cross) then include this region
!                  for the mod-8 correction. Otherwise the region will remain uncorrected in 
!                  order to prevent the infinite cycling of the task.
! 2011-05-21 (VNY) Protecting against possible subscript (out-of-range) errors when working
!                  with the pixMap array (SSC-SPR-6656)
! 2011-06-07 (VNY) Protected against the possibility of getting an undefined value of mod8ymin 
!                  variable
! 2011-07-12 (VNY) Further protection against possible undefined values of mod8ymin
! 2012-12-13 (VNY) Protected from possible subscript errors in the arrays mod8Tile and TotalInMod8Col  

module ommodmap_mod

  use dal
  use param  
  use errorHandling
  use sasdal
  use recover16bits_mod

  implicit none 
  
  private 

  public ommodmap

contains 
  
  !---------------------------------------------------------------------
  subroutine ommodmap
    !---------------------------------------------------------------------
    type(DataSetT) :: inDataSetH                                      ! Input dataset handle.
    type(DataSetT) :: PPSMod8DataSetH                                 ! PPS Modulo-8 Map dataset handle.
    type(DataSetT) :: outDataSetH                                     ! Output dataset handle.
    type(ArrayT) :: inArrayH                                          ! Input array handle.
    type(ArrayT) :: exposureArrayH                                          ! Input exposure array handle.
    type(ArrayT) :: qualArrayH                                        ! Quality Array handle.
    type(ArrayT) :: PPSMod8ArrayH                                     ! PPS Modulo-8 Map array handle.
    type(ArrayT) :: PPSMod8intArrayH                                     ! PPS Modulo-8 Map array handle.
    type(ArrayT) :: outArrayH                                         ! Output array handle.
!    type(TableT) :: inTableH                                          ! Input table handle.
!    type(ColumnT), dimension(4) :: inColumnH                          ! Input image column handle.
    type(AttributableT) :: sourceH                                    ! Source attributable handle.
    type(AttributableT) :: objectH                                    ! Object attributable handle.
    type(BlockT) :: blockH                                            ! Block handle.
    type(ArrayT) :: qualityArrayH
    type(TableT) :: outModesH
    type(ColumnT) :: outColumnH
!    type(AttributeT) :: attributeH
    character(len=255) :: charAttrValue

    character(len=255) :: set                                         ! OM Image Filename.
    character(len=255) :: PPSMod8Set                                  ! OM PPS Modulo-8 Map filename.
    character(len=255) :: outSet                                      ! OM Modulo-8 corrected Image filename.
    character(len=8) :: dataMode                                      ! Input file data mode.
    character(len=255)  :: flatSet                                    ! Input file with the osw flat field
    character(len=255) :: outFlatSet                                  ! file name for the output flat field

    ! variables for the flat image
    type(DataSetT)     :: flatSetH
    type(ArrayT)       :: flatArrayH
    type(DataSetT)     :: outFlatSetH
    type(ArrayT)       :: outFlatArrayH

    integer                      :: flatNdim           ! no of dimensions for the flat array
    integer                      :: flat_nx, flat_ny   ! x, y size of flat field
    integer                      :: flat_dim(2)        ! flat  dimension array
    integer                      :: exposure_nx, exposure_ny   ! x, y size of exposure image
    integer                      :: exposure_dim(2)        ! exposure  dimension array
    real(kind=Single), pointer   :: flatImage(:, :)        ! flat image
    real(kind=Single), pointer   :: exposureImage(:, :)        ! exposure image

    real(kind=double) :: averExposure, sumPixels
    real(kind=single) :: auxAver

    integer :: image_dim(2), quality_dim(2)

    integer(kind=int32) :: nSig                                       ! Significance level for sigma clipping.
    integer(kind=int32) :: nBox                                       ! Size of sliding box (in units of 8 pixels).
!    integer(kind=int32), dimension(:), pointer :: windowX0P           ! WindowX0 pointer.
!    integer(kind=int32), dimension(:), pointer :: windowY0P           ! WindowY0 pointer.
!    integer(kind=int32), dimension(:), pointer :: windowDXP           ! WindowDX pointer.
!    integer(kind=int32), dimension(:), pointer :: windowDYP           ! WindowDY pointer.
    integer(int8), dimension(:, :), pointer :: qualArrayPint8        ! Quality array pointer.
    integer(int16), dimension(:, :), pointer :: qualArrayPint16        ! Quality array pointer.

    integer(int16), dimension(:, :), pointer :: outQualityArrayP  ! Quality-2 array pointer.
    integer(int32), dimension(:), pointer :: outColumnP
    integer(int32), dimension(1:1)  ::  outMode

    integer(int32) :: binAx1                                     ! DPU x-axis binning factors.
    integer(int32) :: binAx2                                     ! DPU y-axis binning factors.
    integer(int32) :: binX                                       ! Bin x.
    integer(int32) :: binY                                       ! Bin y.
    integer(int32) :: windowX0                                   ! WindowX0.
    integer(int32) :: windowY0                                   ! WindowY0.
    integer(int32) :: windowDX                                   ! WindowDX.
    integer(int32) :: windowDY                                   ! WindowDY.
    character(255) :: window0                                   ! Origin of the OSW
    integer(int32) :: windowX1, windowY1   ! Coordinates of the OSW origin
    character(1) :: ch1, ch2, ch3, ch4, ch5

    integer(int32) :: xMin                                       ! Minimum x value for window.
    integer(int32) :: xMax                                       ! Maximum x value for window.
    integer(int32) :: yMin                                       ! Minimum y value for window.
    integer(int32) :: yMax                                       ! Maximum y value for window.
    integer(int32) :: mod8xmin                                   ! Minimum x value for window.
    integer(int32) :: mod8xmax                                   ! Maximum x value for window.
    integer(int32) :: mod8ymin                                   ! Minimum y value for window.
    integer(int32) :: mod8ymax                                   ! Maximum y value for window.
    integer(int32) :: allocErr                                   ! Allocate error status.
    integer(int32) numPixels                                     ! Number of pixels.
    integer(int32) :: x                                          ! X.
    integer(int32) :: y                                          ! Y.
    integer(int32) :: i,i1,i2                                          ! Count.
    integer(int32) :: j,j1,j2                                          ! Count.
    integer(int32) :: n                                          ! Count.
    integer(int32) :: mod8TileDX                                 ! Modulo-8 tile DX.
    integer(int32) :: mod8TileDY                                 ! Modulo-8 tile DY.
    integer(int32) :: boxDX                                      ! Box DX.
    integer(int32) :: boxDY                                      ! Box DY.
    integer(int32) :: minDX                                      ! Minimum DX.
    integer(int32) :: minDY                                      ! Minimum DY.
    integer(int32) :: maxDX                                      ! Maximum DX.
    integer(int32) :: maxDY                                      ! Maximum DY.
    integer(int32) :: RunningPixel                               ! Running Pixel index
    integer(int32) :: nBoxTimesmod8TileDX                        ! nbox * mod8TileDX
    integer(int32) :: nBoxTimesmod8TileDY                        ! nbox * mod8TileDY

!    real(kind=single), dimension(:, :), pointer :: inOSWP             ! Input observation science window pointer.
    real(single), dimension(:, :), pointer :: PPSMod8P           ! PPS Modulo-8 Map pointer.
    integer(int32), dimension(:, :), pointer :: PPSMod8intP           ! PPS Modulo-8 Map pointer.
    real(single), dimension(:, :), pointer :: outOSWP            ! Output observation science window pointer.
!    integer(kind=int32), dimension(:, :), pointer :: outOSWintP            ! Output observation science window pointer.
    real(single), dimension(:, :), pointer :: outFlatP            ! Output flat field  pointer (real).
    integer(int32), dimension(:, :), pointer :: outFlatIntP            ! Output flat field pointer (integer).
    real(single), allocatable, dimension(:, :) :: inOSW          ! Input observation science window.
    real(single) :: mean                                         ! Mean.
    real(single) :: sigma                                        ! Sigma.
    real(single), allocatable, dimension(:, :) :: mod8Map        ! Modulo-8 map.
    real(single), allocatable, dimension(:, :) :: mod8Tile       ! Modulo-8 tile.
    real(single), allocatable, dimension(:, :) :: meanTile       ! Mean tile.
    real(single), allocatable, dimension(:, :) :: outOSW         ! Output observation science window.
    real(single), allocatable, dimension(:) :: TotalInMod8Col    ! Sum over mod 8 cell columns of mod-8 map
    real(single), allocatable, dimension(:) :: TotalInOSWCol     ! Sum over mod 8 cell columns of input OSW
    real(single), allocatable, dimension(:) :: TotalInMod8Row    ! Sum over mod 8 cell rows of mod-8 map
    real(single), allocatable, dimension(:) :: TotalInOSWRow     ! Sum over mod 8 cell rows of input OSW
    real(single), allocatable, dimension(:, :) :: StartImTile    ! starting image tile.
    real(single), allocatable, dimension(:, :) :: InterImTile    ! Intermediate image tile.
    real(single) :: TotalInMod8Cols                              ! Total of all Mod 8 Cols
    real(single) :: TotalInMod8Rows                              ! Total of all Mod 8 Rows
    real(single) :: RunningMod8Total                             ! Running total of mod-8 values
    real(single) :: RunningPosition                              ! Current position of virtual pixel boundary
    real(single) :: OldRunningPosition                           ! Position of previous virtual pixel boundary
    real(single) :: tol                                          ! tolerance
    real(single) :: realMod8TileDX                               ! Stores real(mod8TileDX)
    real(single) :: realMod8TileDY                               ! Stores real(mod8TileDY)
    real(single) :: realRunningPixel                             ! stores real(RunningPixel)
    logical(bool) :: multiplyFlat                                ! Flag for multiplication by the flat field
                                                                      ! instead of redistribution of the counts
    logical(bool) :: applyFlat

    logical(bool) :: PPSProduct                                  ! PPS Product flag.
    logical(bool) :: mod8correction                              ! modulo-8 correction flag.
    integer(int32) :: mod8correctionPar                          ! parameter for modulo-8 correction
    logical(bool) :: binBPE                                      ! BPE binning flag.
    logical(bool), dimension(:, :), allocatable :: qualArray     ! Quality array.
    integer(int16), dimension(:, :), allocatable :: outQualArray 

    logical(bool), dimension(:, :), allocatable :: pixMap        ! Pixel map.
    logical(bool), dimension(:, :), allocatable :: pixMask      ! Pixel map.
    logical(bool), dimension(:, :), allocatable :: tempPixMask1   ! temporaryPixel map.
    logical(bool), dimension(:, :), allocatable :: tempPixMask2   ! temporaryPixel map.
    logical(bool), dimension(:, :), allocatable :: tempPixMask3   ! temporaryPixel map.

    logical(bool) :: auxMask3
    integer :: sizeMask3filter

    character(6)            :: omFilter   ! the OM filter
!    character(len=8)            :: omMode     ! FAST or IMAGING mode 
    integer :: winX0, winY0, winDX, winDY
    logical(bool) :: hasOMIMI
    logical(bool) :: hasPrimary !for the flat field
    logical(bool) :: hasQUALITY
    logical(bool) :: hasEXPOSURE
    logical(bool) :: usingFlat
    logical(bool) :: produceOutputFlat
    real(single), allocatable, dimension(:, :) :: partialFlat ! a part of the flat field corresponding to OSW 
    real(single) :: partialMean, partialSum
    integer(int32) :: ii,jj
    integer(int32) :: ii1,jj1
    integer(int32) :: countAux

    real(Double) :: sumQuality

    logical(bool) :: auxlogical
    !---------------------------------------------------------------
    !
    ! AMK
    !
    integer           :: istat
    !
    !--------------------------------------------------------------
    !integer                      :: ndim     ! no of dimensions
    integer                      :: nx, ny   ! x,ysize of image
    integer                      :: dim(2)   ! image dimension array

    integer(int32) :: outDim(2) 
    integer(int32), pointer :: int32array(:, :)
    real(Single), pointer :: re32array(:, :)
    !real(kind=Single), pointer   :: image(:, :)
    real(Double) :: meanCheck, sumCheck, normFlat
    character(255) :: detsec, detsize

    integer :: iomFilter
    integer :: qdatatype

    sizeMask3filter=30


    !*********************************************************
    ! 2011-07-12 initialising the values of mod8xmin, 
    ! mod8xmax, mod8ymin, mod8ymax to protect against possible
    ! crashes 
    !********************************************************
    mod8xmin=0
    mod8xmax=0
    mod8ymin=0
    mod8ymax=0

    produceOutputFlat=.false.
    applyFlat = .false. 
    multiplyFlat = .false.
    ! Read in parameters.
    set = stringParameter("set")
    PPSProduct = booleanParameter("mod8product")
    PPSMod8Set = stringParameter("mod8set")
    !applyFlat = booleanParameter("applyflat")
    !multiplyFlat = booleanParameter("multiplyflat")
    outSet = stringParameter("outset") 
    nSig = intParameter("nsig")
    nBox = intParameter("nbox")
    mod8correctionPar = intParameter("mod8correction")
    mod8correction = .false.
    if (mod8correctionPar .gt. 0) mod8correction = .true.
    if (mod8correctionPar .eq. 2)  applyFlat = .true.
    if (mod8correctionPar .eq. 3) multiplyFlat = .true.

    if (applyFlat) then
       flatSet= stringParameter("flatset")
    endif
    if (flatSet .eq. "") then
       call warning("The use of a flat field requested but no filename provided")
       applyFlat=.false.
    endif


    ! check parameter consistency
    if (PPSProduct .and. PPSMod8Set .eq. "-") then
       call fatal("paramError","Mod8 product requested but no filename provided")
    end if
    outFlatSet= stringParameter("outflatset")
    if (outFlatSet .eq. "") then 
       produceOutputFlat = .false.
    else
       produceOutputFlat= .true.
    endif

    usingFlat=.false.
    if (applyFlat .and. flatSet .ne. "") then
       flatSetH=dataSet(flatSet, read)
       usingFlat=.true.
    else
       usingFlat=.false.
    endif
!**************************************************************
! Dealing with the flat image
    if (usingFlat .and.  mod8correction) then
       hasPrimary = .false. 
       if (hasBlock(flatSetH, "PRIMARY")) then
          hasPrimary= .true.
       endif

       if (hasPrimary) then
          flatArrayH = array(flatSetH, "PRIMARY", read) ! With int32 data
       endif

       flatNdim  = numberOfDimensions(flatArrayH)             ! How many dimensions?

       ! If it's not an image then crash out
       if (flatNdim .ne. 2) call error('notImage','Flat filed  file is not an image')

       flat_dim = dimensions(flatArrayH)               ! Get the size of the image
       flat_nx = flat_dim(1)
       flat_ny = flat_dim(2)

       if (hasPrimary) then
          re32array=> real32array2data(flatArrayH)
          !    else
          !       int32array=> int32array2data(flatArrayH)
       endif

       allocate(flatImage(0 : flat_nx - 1, 0 : flat_ny - 1), stat = allocErr)
       call CheckAllocation(allocErr, "flatImage")
       flatImage=0.0

       if (hasPrimary) then
          flatImage=re32array
       endif
    endif ! usingFlat .and. mod8noiseCorrection

    ! Read in the OM OSW Image file.
    inDataSetH = dataSet(set, read)

    ! Determine file type from datamode in primary header.
    ! dataMode = stringAttribute(inDataSetH, "DATAMODE")
    dataMode = "IMAGING"

    binBPE = .false.
    binAx1 = 0
    binAx2 = 0

    hasOMIMI= .false.
    if (hasBlock(inDataSetH, "OMIMI1")) then
       hasOMIMI= .true.
    endif

    hasQUALITY= .false.
    if (hasBlock(inDataSetH, "QUALITY")) then
       hasQUALITY= .true.
    endif

    hasEXPOSURE= .false.
    if (hasBlock(inDataSetH, "EXPOSURE")) then
       hasEXPOSURE= .true.
       exposureArrayH = array(inDataSetH, "EXPOSURE", read)
    endif

    if (hasOMIMI) then  
       inArrayH = array(inDataSetH, "OMIMI1", read) ! With int32 data
    else
       inArrayH = array(inDataSetH, "PRIMARY", read) ! With float32 data
    endif

    if (hasOMIMI) then
       binBPE = booleanAttribute(inArrayH, "BINBPE")
       binAx1 = int32Attribute(inArrayH, "BINAX1")
       binAx2 = int32Attribute(inArrayH, "BINAX2")
       if (hasAttribute(inArrayH, "FILTER")) then
          omFilter = stringAttribute(inArrayH, "FILTER")
       else
          omFilter="NONE"
       endif
       ! VNY 26.11.2003
       image_dim = dimensions(inArrayH)

       ! Calculate image binning.
       binX=binAx1
       binY=binAx2
 

       if (trim(omFilter) .eq. "GRISM1" .or. trim(omFilter) .eq. "Grism1" .or. &
           trim(omFilter) .eq. "grism1" .or. trim(omFilter) .eq. "GRISM2" .or. &
           trim(omFilter) .eq. "Grism2" .or. trim(omFilter) .eq. "grism2") then

           if (binAx1 .eq. 0) binX=1
           if (binAx2 .eq. 0) binY=1
           if (binBPE) then
              binX=2
              binY=2
           endif
       else ! NON-GRISM FILTERS
           binX = 2 ** binAx1
           binY = 2 ** binAx2
           if (binBpe) then
               binX = 2 * binX
               binY = 2 * binY
           end if 


       end if


       sizeMask3filter=sizeMask3filter/binX

       if (hasAttribute(inArrayH,"WINDOWX0")) then
          windowX0= int32Attribute(inArrayH, "WINDOWX0")
          windowY0= int32Attribute(inArrayH, "WINDOWY0")
          windowDX= int32Attribute(inArrayH, "WINDOWDX")
          windowDY= int32Attribute(inArrayH, "WINDOWDY")


          if (trim(omFilter) .eq. "GRISM1" .or. trim(omFilter) .eq. "Grism1" .or. &
               trim(omFilter) .eq. "grism1" .or. trim(omFilter) .eq. "GRISM2" .or. &
               trim(omFilter) .eq. "Grism2" .or. trim(omFilter) .eq. "grism2") then


             winX0=windowX0/binX
             winY0=windowY0/binY
             winDX=windowDX/binX
             winDY=windowDY/binY
          else
             winX0=windowX0
             winY0=windowY0
             winDX=windowDX
             winDY=windowDY
          end if

       else
          if (hasAttribute(inArrayH,"DETSEC")) then
             window0 = stringAttribute(inArrayH, "DETSEC")
             read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
                  ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5
             winX0=windowX0
             winY0=windowY0
             winDX=windowX1-windowX0+1
             winDY=windowY1-windowY0+1
             ! VNY 26.11.2003
             image_dim=dimensions(inArrayH)
             if (winDX .ne. image_dim(1)) then
                write (errstr, *) "Image X-size in DETSEC-attribute  is not correct"
                call warning("Size")
                winDX=image_dim(1)
                windowX1=windowX1-1
             endif
             if (winDY .ne. image_dim(2)) then
                write (errstr, *) "Image Y-size in DETSEC-attribute is not correct"
                call warning("Size")
                winDY=image_dim(2)
                windowY1=windowY1-1
             endif
          else
             dim = dimensions(inArrayH)   ! Get the size of the image 
             winX0=0
             winY0=0
             !winX0=1
             !winY0=1
             winDX = dim(1)
             winDY = dim(2)
          endif
       endif
    else
       binAx1 = int32Attribute(inDataSetH, "BINAX1")
       binAx2 = int32Attribute(inDataSetH, "BINAX2")
       binBPE = booleanAttribute(inDataSetH, "BINBPE")
       if (hasAttribute(inDataSetH, "FILTER")) then
          omFilter = stringAttribute(inDataSetH, "FILTER")
       else
          omFilter="NONE"
       endif
       ! Calculate image binning.
       ! VNY 26.11.2003
       image_dim = dimensions(inArrayH)
       binX=binAx1
       binY=binAx2

       if (trim(omFilter) .eq. "GRISM1" .or. trim(omFilter) .eq. "Grism1" .or. &
           trim(omFilter) .eq. "grism1" .or. trim(omFilter) .eq. "GRISM2" .or. &
           trim(omFilter) .eq. "Grism2" .or. trim(omFilter) .eq. "grism2") then

           if (binAx1 .eq. 0) binX=1
           if (binAx2 .eq. 0) binY=1
           if (binBPE) then
              binX=2
              binY=2
           endif
       else ! NON-GRISM FILTERS
           binX = 2 ** binAx1
           binY = 2 ** binAx2
           if (binBpe) then
               binX = 2 * binX
               binY = 2 * binY
           end if 


       end if


       if (hasAttribute(inDataSetH,"WINDOWX0")) then
          windowX0= int32Attribute(inDataSetH, "WINDOWX0")
          windowY0= int32Attribute(inDataSetH, "WINDOWY0")
          windowDX= int32Attribute(inDataSetH, "WINDOWDX")
          windowDY=int32Attribute(inDataSetH, "WINDOWDY")



          if (trim(omFilter) .eq. "GRISM1" .or. trim(omFilter) .eq. "Grism1" .or. &
               trim(omFilter) .eq. "grism1" .or. trim(omFilter) .eq. "GRISM2" .or. &
               trim(omFilter) .eq. "Grism2" .or. trim(omFilter) .eq. "grism2") then



             winX0=windowX0/binX
             winY0=windowY0/binY
             winDX=windowDX/binX
             winDY=windowDY/binY

          else

             winX0=windowX0
             winY0=windowY0
             winDX=windowDX
             winDY=windowDY
          end if


       else
          if (hasAttribute(inDataSetH,"DETSEC")) then
             window0 = stringAttribute(inDataSetH, "DETSEC")
            read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
                 ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5
             winX0=windowX0
             winY0=windowY0
             winDX=windowX1-windowX0+1
             winDY=windowY1-windowY0+1
             ! VNY 26.11.2003 Check the consistency of the image dimensions
             image_dim=dimensions(inArrayH)
             if (winDX .ne. image_dim(1)) then
                write (errstr, *) "Image X-size in DETSEC-attribute is not correct"
                call warning("Size")
                winDX=image_dim(1)
                windowX1=windowX1-1
             endif
             if (winDY .ne. image_dim(2)) then
                write (errstr, *) "Image Y-size in DETSEC-attribute is not correct"
                call warning("Size")
                winDY=image_dim(2)
                windowY1=windowY1-1
             endif
          else
             dim = dimensions(inArrayH)   ! Get the size of the image 
             winX0=0
             winY0=0
             !winX0=1
             !winY0=1
             winDX = dim(1)
             winDY = dim(2)
          endif
       endif
    endif ! has OMIMI


    ! Calculate size of modulo-8 tile.
    mod8TileDX = 8 / binX
    mod8TileDY = 8 / binY
    realMod8TileDX = real(mod8TileDX)
    realMod8TileDY = real(mod8TileDY)


    ! Allocate space for the quality array.
    ! Could be a problem here if binX and binY are not 1 and windowDX and windowDY are not modulo 8
    !

!**************************************************
!
!***************************************************
  
    windowX0=winX0
    windowY0=winY0
    windowDX=winDX
    windowDY=winDY

    !&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

    ! VNY 03.10.2003 Dividing by binX and binY may cause problems: it might contradict to the real image size
    !xMin = 1
    !xMax = windowDX / binX
    !yMin = 1
    !yMax = windowDY / binY
    xMin = windowX0
    xMax = xMin+windowDX - 1
    yMin = windowY0
    yMax = yMin + windowDY - 1

    write(*,*) "L611 xMin=", xMin, " xMax=", xMax, " yMin=", yMin, " yMax=", yMax

    !--- Allocating the partial flat image (sized to the OSW)
    if (mod8correction) then 
       allocate(partialFlat(xMin:xMax, yMin:yMax), stat=allocErr)
       call CheckAllocation(allocErr, "partialFlat")
       !---Fill in the partial image
       partialFlat=1.0
       if (usingFlat) then
          do i=xMin, xMax
             do j=yMin, yMax
                partialMean=0.0
                partialSum=0.0
                do ii=0,binX-1
                   do jj=0,binY-1
                      if ((i)*binX+ii .le. 2047 .and. (j)*binY+jj .le. 2047) then 
                         partialMean=partialMean+flatImage((i)*binX+ii, (j)*binY+jj) 
                         partialSum=partialSum+1.0
                      endif
                   enddo
                enddo
                if (partialSum .gt. 0.0) partialMean=partialMean/partialSum
                partialFlat(i,j)=partialMean
             enddo
          enddo
       endif
    endif ! mod8correction


    allocate (qualArray(xMin : xMax, yMin : yMax), stat = allocErr)
    call CheckAllocation(allocErr, "qualArray")

    allocate (outQualArray(xMin : xMax, yMin : yMax), stat = allocErr)
    call CheckAllocation(allocErr, "outQualArray")

    ! Checking the number of bad pixels marked by the quality array
    allocate (pixMap(xMin : xMax, yMin : yMax), stat = allocErr)
    !write(*,*) "L648 xMin=", xMin, " xMax=", xMax, " yMin=", yMin, " yMax=", yMax
    call CheckAllocation(allocErr, "pixMap")
    ! Fill the pixel map.
    pixMap = .true.
    sumQuality=count(pixMap)

    ! Deallocate the temporary pixMasks
    deallocate(pixMap, stat=allocErr)
    call CheckDeallocation(allocErr, "pixMap")


    outQualArray=0

    ! Get a handle and pointer on the quality array and extract the quality data.
    qualArray = .false.
    if (hasQUALITY) then
       qualArrayH = array(inDataSetH, "QUALITY", read)
       qdatatype = datatype(qualArrayH)

       if (qdatatype .eq. 2) then

          qualArrayPint8 => int8Array2Data(qualArrayH)  
          outQualArray = int(qualArrayPint8, int16)
          where (qualArrayPint8 .ne. 0) 
             qualArray = .true.
          elsewhere
             qualArray = .false.
          end where
       else
          qualArrayPint16 => int16Array2Data(qualArrayH)
          outQualArray = qualArrayPint16
          where (qualArrayPint16 .ne. 0) 
             qualArray = .true.
          elsewhere
             qualArray = .false.
          end where
       endif
       !qualArrayP => int8Array2Data(qualArrayH)
       !outQualArray=qualArrayP
       !where (qualArrayP .ne. 0) 
       !   qualArray = .true.
       !elsewhere
       !   qualArray = .false.
       !end where
    endif

    ! Check the number of bad pixels marked in the quality array
    sumCheck=count(qualArray)

    if (sumCheck .gt. sumQuality*0.6) then
       write (errstr, *) "Too many (more than 50%) pixels in the quality array are marked as bad."
       call warning("QualityMap")
       qualArray=.false.
    endif
            
    ! Allocate space for the input observation science window.
    allocate (inOSW(xMin : xMax, yMin : yMax), stat = allocErr)
    call CheckAllocation(allocErr, "inOSW")

    ! Get a handle and pointer on the input observation science window array and extract the OSW data.
    if (hasOMIMI) then
       inArrayH = array(inDataSetH, "OMIMI1", read) ! With int32 data
    else
       inArrayH = array(inDataSetH, "PRIMARY", read) ! With float32 data
    endif

    dim = dimensions(inArrayH)               ! Get the size of the image       
    nx = dim(1)
    ny = dim(2)
    outDim=dim
       

    sumPixels=0.0_double
    averExposure=0.0_double
    if (hasEXPOSURE) then
       re32array=> real32array2data(exposureArrayH)
       exposure_dim = dimensions(exposureArrayH)               ! Get the size of the image
       exposure_nx = exposure_dim(1)
       exposure_ny = exposure_dim(2)
       allocate(exposureImage(0 : exposure_nx - 1, 0 : exposure_ny - 1), stat = allocErr)
       call CheckAllocation(allocErr, "exposureImage")
       exposureImage=0.0
       exposureImage=re32array

       i1=exposure_nx/2
       j1=exposure_ny/2
       auxAver=exposureImage(i1,j1)*0.8
       do i1=0,exposure_nx-1
          do j1=0,exposure_ny-1
             if (exposureImage(i1,j1) .gt. auxAver) then
                averExposure=averExposure+exposureImage(i1,j1)
                sumPixels=sumPixels+1.0_double
             endif
          enddo
       enddo
       if (sumPixels .gt. 1.0_double) averExposure=averExposure/sumPixels
    endif


    if (hasOMIMI) then
       int32array=> int32array2data(inArrayH)
    else
       re32array=> real32array2data(inArrayH)
    endif

    !*****************************************************
    ! Fill in the working array with the data
    inOSW=0.0 
    if (hasOMIMI) then
       inOSW = real(int32array)
    else
       inOSW=re32array
    endif
    
    if (hasEXPOSURE) then
       where(exposureImage .gt. 0.0) inOSW=inOSW*exposureImage
    endif
      
    !*******************************************************
    ! Release handles on data.
    call release(inArrayH)
    if (hasQUALITY) then
       call release(qualArrayH)
    endif

    ! Allocate space for the output observation science window.
    allocate (outOSW(xMin : xMax, yMin : yMax), stat = allocErr)
    call CheckAllocation(allocErr, "outOSW")

    ! MJP because sets image outside mod 8 pixels with twice input values since /2 later
    if (mod8correction) then 
       outOSW = inOSW * 2.0
 
       call recover16bits(xMin,xMax, yMin, yMax, inOSW, binX, binY)
 
       !
 
       !&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& 
       ! Allocate space for the pixel map.
       allocate (pixMap(xMin : xMax, yMin : yMax), stat = allocErr)
       !write(*,*) "L788 xMin=", xMin, " xMax=", xMax, " yMin=", yMin, " yMax=", yMax
       call CheckAllocation(allocErr, "pixMap")
       ! Fill the pixel map.
       pixMap = .true.
       sumCheck=count(pixMap)


      ! VNY Checking for the mean values
       meanCheck=sum(outOSW)
       if (sumCheck .gt. 0.0) meanCheck=meanCheck/sumCheck

       meanCheck=sum(inOSW)      
       if (sumCheck .gt. 0.0) meanCheck=meanCheck/sumCheck


       meanCheck=sum(partialFlat)
       if (sumCheck .gt. 0.0) meanCheck=meanCheck/sumCheck
 

       normFlat=1.0
       if (meanCheck .gt. 0.0) normFlat=meanCheck
 
      ! This can be used instead of distribution of counts 
       if (multiplyFlat .and. usingFlat) then
          do i=xMin, xMax
             do j=yMin, yMax
                if (partialFlat(i,j) .ne. 0.0) then
                   ! VNY 29.10.2003 Normalising using the  flat field (actually, the normalisation gives nothing)
                   outOSW(i,j)= inOSW(i,j) * partialFlat(i,j)
!                   outOSW(i,j)= inOSW(i,j) * partialFlat(i,j)/normFlat
                   !outOSW(i,j)= outOSW(i,j) * partialFlat(i,j)/normFlat
                endif
                if (inOSW(i,j) .gt. 50000.) then
                   write (errstr, *) "The pixel i=",i, " j=",j, " is unrealistically bright"
                   call warning("InputOSW")
                endif
             enddo
          enddo
       endif


       meanCheck=sum(inOSW)
       if (sumCheck .gt. 0.0) meanCheck=meanCheck/sumCheck

       if (usingFlat .and. mod8correction) then
          partialFlat=partialFlat*meanCheck
       endif
       ! Deallocate the temporary pixMasks
       deallocate(pixMap, stat=allocErr)
       call CheckDeallocation(allocErr, "pixMap")

       meanCheck=sum(outOSW)
       if (sumCheck .gt. 0.0) meanCheck=meanCheck/sumCheck

       if (.not. multiplyFlat) then 
          !
          ! Deal with the window sizes so that the correct pixels are processed for the mod 8 noise.
          !
          !
          ! Calculate size of box.
          boxDX = mod8TileDX * floor(nBox / 2.0)
          boxDY = mod8TileDY * floor(nBox / 2.0)
          !write(*,*) "L850 mod8TileDX=", mod8TileDX, " mod9TileDY=", mod8TileDY
          !write(*,*) "L851 boxDX=", boxDX, " boxDY=", boxDY

          ! 2011-06-07 protection against undefined mod8xmin
          mod8xmin = xmin
          if (mod(((windowX0) + xmin),mod8TileDX) .eq. 0)then
             mod8xmin = xmin
          else
             mod8xmin = (((xmin + (windowX0))/mod8TileDX) + 1)
             mod8xmin = mod8TileDX * mod8xmin - (windowX0) + 1
          endif
          if (mod8xmin .lt. xmin) mod8xmin=xmin

          ! VNY 09.06.2004 Here is a problem: xmax is understood here as
          ! the inctrement of the window but not as the maximal 
          ! pixel number

          ! 2011-06-07 protection against undefined mod8xmax
          mod8xmax = xmax
          if (mod(((windowX0) + xmax),mod8TileDX) .eq. 0)then
             mod8xmax = xmax
          else
             mod8xmax = mod8TileDX * (((xmax + (windowX0))/mod8TileDX)) - (windowX0)
          endif
          if (mod8xmax .gt. xmax-1) mod8xmax=xmax-1

          ! 2011-06-07 protection against undefined mod8ymin
          mod8ymin = ymin
          if (mod(((windowY0) + ymin),mod8TileDY) .eq. 0)then
             mod8ymin = ymin
          else
             mod8ymin = (ymin+ (windowY0)) / mod8TileDY
             mod8ymin = (((ymin+ (windowY0)) / mod8TileDY) + 1)
             mod8ymin = mod8TileDY * mod8ymin - (windowY0) + 1
          endif
          if (mod8ymin .lt. ymin) mod8ymin=ymin

          ! 2011-06-07 protection against undefined mod8ymax
          mod8ymax = ymax
          if (mod(((windowY0) + ymax),mod8TileDY) .eq. 0)then
             mod8ymax = ymax
          else
             mod8ymax = mod8TileDY * (((ymax + (windowY0))/mod8TileDY)) - (windowY0)
          endif
          if (mod8ymax .gt. ymax-1) mod8ymax=ymax-1

          !************************************************************
          write(*,*) "L884 mod8xmin=", mod8xmin, " mod8xmax=", mod8xmax
          write(*,*) "L885 mod8ymin=", mod8ymin, " mod8ymax=", mod8ymax
          !************************************************************
          ! Check to see if modulo-8 noise is lost in binning.
          if (.not. ((binX >= 8) .and. (binY >= 8))) then
             ! Allocate space for the pixel map.
             allocate (pixMap(xMin : xMax, yMin : yMax), stat = allocErr)
             !write(*,*) "L888 xMin=", xMin, " xMax=", xMax, " yMin=", yMin, " yMax=", yMax
             call CheckAllocation(allocErr, "pixMap")
 
             ! Fill the pixel map.
             pixMap = .true.
             where (qualArray) pixMap = .false.

             ! Free up memory.
             deallocate(qualArray, stat=allocErr)
             call CheckDeallocation(allocErr, "qualArray") 

             ! Check the image has enough counts to be corrected for mod8 noise
             allocate (tempPixMask1(xMin : xMax, yMin : yMax), stat = allocErr)
             call CheckAllocation(allocErr, "tempPixMask")
          
             allocate (tempPixMask2(xMin : xMax, yMin : yMax), stat=allocErr)
             call CheckAllocation(allocErr, "tempPixMask2")

             allocate (tempPixMask3(xMin : xMax, yMin : yMax), stat=allocErr)
             call CheckAllocation(allocErr, "tempPixMask3")

             tempPixMask1 = .false.  ! Sets entire array to zero
             tempPixMask2 = pixMap   ! Sets tempPixMask2 equal to pixMap
             tempPixMask3 = pixMap   ! Sets tempPixMask3 equal to pixMap
         
             !*********************************************************************
             ! So tempPixMask2 is true if quality data is zero, false otherwise
             !
             ! This returns the number of elements that are set to true in each array
             !
             !**********************************************************************

             !*******************************************************************
             ! VNY 04.04.2008 The use of tempPixMask3 is designed for avoiding 
             ! the infinite cycling of the task when building the modulo-8 map.
             ! However, small areas with bad pixels now have unpleasant black pixels
             ! in some (not all) images.
             ! To improve the situation we can return to the previous algorithm wherever
             ! the bad pixel areas are not very large. So, we can filter the tempPixMask3
             !***************************************************************************

             ! Filtering the mask3
             do i=xMin, xMax
                do j=yMin, yMax
                   auxMask3=.false.
                   do ii=i-sizeMask3filter,i+sizeMask3filter
                      if (ii .ge. xMin .and. ii .le. xMax) then
                         if (tempPixMask2(ii,j)) auxMask3=.true.
                      endif
                   enddo
                   do jj=j-sizeMask3filter,j+sizeMask3filter
                      if (jj .ge. yMin .and. jj .le. yMax) then
                         if (tempPixMask2(i,jj)) auxMask3=.true.
                      endif
                   enddo
                   tempPixMask3(i,j)=auxMask3
                enddo
             enddo

             countAux=0
             do while (count(tempPixMask1) .ne. count(tempPixMask2)) 
                ! Do this while number of elements that are true in each Array are not equal
                tempPixMask1 = tempPixMask2
                numPixels = count(tempPixMask1)
                countAux=countaux+1

                !******************************************************************
                if (usingFlat) then
                   mean = sum(partialFlat, mask=tempPixMask1) / real(numPixels) !
                   sigma = sqrt(sum(((partialFlat - mean) ** 2),& 
                        mask=tempPixMask1) / real(numPixels - 1))
                   where (abs(partialFlat - mean) / sigma > nSig)& 
                        tempPixMask2 = .false.
                else
                   mean = sum(inOSW, mask=tempPixMask1) / real(numPixels)
                   ! This sums those elements of inOSW for which tempPixMask1 
                   ! has corresponding element that is true - that is zero
                   sigma = sqrt(sum(((inOSW - mean) ** 2),& 
                        mask=tempPixMask1) / real(numPixels - 1))
                   where (abs(inOSW - mean) / sigma > nSig)& 
                        tempPixMask2 = .false.
                endif
                !***************************************************************** 
             end do

             ! Deallocate the temporary pixMasks
             deallocate(tempPixMask1, stat=allocErr)
             call CheckDeallocation(allocErr, "tempPixMask1")

             deallocate(tempPixMask2, stat=allocErr)
             call CheckAllocation(allocErr, "tempPixMask2")

             ! Check for a very high values of pixels, which might provoke an infinite cycling 
             ! of the programme
             do i=xMin, xMax
                do j=yMin, yMax
                   if (inOSW(i,j) .gt. 250000.) then
                      write (errstr, *) "The pixel i=",i, " j=",j, " is unrealistically bright"
                      call warning("InputOSW")
                      !inOSW(i,j)=0.0
                   endif
                enddo
             enddo
 
             ! Is the mean countrate big enough?
             ! mod8 noise up to 10-20% effect, assume that must have ~3% accuracy to improve image
             ! this corresponds approx to mean * nBox^2 > 1000  
             if (mean * real(nBox * nBox) .gt. 1000.0) then

                ! Allocate space for pixel mask.
                allocate (pixMask(xMin : xMax, yMin : yMax), stat = allocErr)
                !write(*,*) "L999 pixMask: xMin=", xMin, " xMax=", xMax, " yMin=", yMin, " yMax=", yMax
                call CheckAllocation(allocErr, "pixMask")

                ! Allocate space for the modulo-8 map.
                allocate (mod8Map(xMin : xMax, yMin : yMax), stat = allocErr)
                call CheckAllocation(allocErr, "mod8Map")
 
                ! Calculate size of box.
                boxDX = mod8TileDX * floor(nBox / 2.0)
                boxDY = mod8TileDY * floor(nBox / 2.0)
                !write(*,*) "L1011 boxDX=", boxDX, " boxDY=", boxDY

                ! Calculate good pixels to be used in producing the mod8 Tile Map 
                ! using a sigma clipping method.
                ! Go through each pixel 1 by 1 to find out if it is a fluctuation with a floating box
                pixMask = .true.
                nBoxTimesmod8TileDX = nBox * mod8TileDX   ! Added to speed things up
                nBoxTimesmod8TileDY = nBox * mod8TileDY   ! added to speed things up
 
                ! NB here other operators rather than where do not work
                !*******************************
                where (inOSW > 250000.) inOSW=0.0
                !^^^^^^^^^^^^^^^^^^^^^^ 
                x=mod(mod8xmax,boxDX)
               
                do x = mod8xmin, mod8xmax, boxDX
                   !write(*,*) "L1025 x=", x
                   do y = mod8ymin, mod8ymax, boxDY
                      minDX = max(mod8xmin, x -  nBoxTimesmod8TileDX)
                      minDY = max(mod8ymin, y -  nBoxTimesmod8TileDY)
                      maxDX = min(mod8xmax, x +  nBoxTimesmod8TileDX)
                      maxDY = min(mod8ymax, y +  nBoxTimesmod8TileDY)
                      !if (y .eq. mod8ymin) then
                      !   write(*,*) "L1033 y=", y, " minDY=", minDY
                      !endif
                      allocate (tempPixMask1(minDX : maxDX, minDY : maxDY), stat = allocErr)
                      call CheckAllocation(allocErr, "tempPixMask1")

                      allocate (tempPixMask2(minDX : maxDX, minDY : maxDY), stat=allocErr)
                      call CheckAllocation(allocErr, "tempPixMask2")

                      tempPixMask1 = .false.

                      !*********************************************************
                      ! 2011-05-21 Protecting the bounds of pixMap from crashing
                      !*********************************************************
                      if (minDX .lt. xMin) minDX=xMin
                      if (minDY .lt. yMin) minDY=yMin
                      if (maxDX .gt. xMax) maxDX=xMax
                      if (maxDY .gt. yMax) maxDY=yMax
                      tempPixMask2 = pixMap(minDX : maxDX, minDY : maxDY)
                      !write(*,*) "L1044 xMin=", xMin, " xMax=", xMax, " yMin=", yMin, " yMax=", yMax
                      !write(*,*) "L1044 minDX=", minDX, " maxDX=", maxDX, " minDY=", minDY, " maxDY=", maxDY
                      do while (count(tempPixMask1) /= count(tempPixMask2))
                         ! VNY there is a risk here to cycle indefinitely
                         tempPixMask1 = tempPixMask2
                         numPixels = count(tempPixMask1)

                         if (usingFlat) then
                            mean = sum(partialFlat(minDX : maxDX, minDY : maxDY), &
                                 mask = tempPixMask1) / real(numPixels)
                            sigma = sqrt(sum(((partialFlat(minDX : maxDX, minDY : maxDY) - mean) ** 2), & 
                                 mask = tempPixMask1) / real(numPixels - 1))
                            where (abs(partialFlat(minDX : maxDX, minDY : maxDY) & 
                                 - mean) / sigma > nSig) & 
                                 tempPixMask2 = .false.
                         else
                            mean = sum(inOSW(minDX : maxDX, minDY : maxDY), &
                                 mask = tempPixMask1) / real(numPixels)
                            sigma = sqrt(sum(((inOSW(minDX : maxDX, minDY : maxDY) - mean) ** 2), & 
                                 mask = tempPixMask1) / real(numPixels - 1))
                            where (abs(inOSW(minDX : maxDX, minDY : maxDY) - mean) / sigma > nSig) & 
                                 tempPixMask2 = .false.
                         endif
                      end do

                      if (usingFlat) then
                         do i = 1, boxDX
                            do j = 1, boxDY
                               if (x + i - 1 <= mod8xmax .AND. y + j - 1 <= mod8ymax) then
                                  if (abs(partialFlat(x + i - 1, y + j - 1) - mean) / sigma > nSig)& 
                                       pixMask(x + i - 1, y + j - 1) = .false.
                               end if
                            end do
                         end do
                      else
                         do i = 1, boxDX
                            do j = 1, boxDY
                               if (x + i - 1 <= mod8xmax .AND. y + j - 1 <= mod8ymax) then
                                  if (abs(inOSW(x + i - 1, y + j - 1) - mean) / sigma > nSig)& 
                                       pixMask(x + i - 1, y + j - 1) = .false.
                               end if
                            end do
                         end do
                      endif
                      ! and deallocate the temporary pixMasks
                      deallocate(tempPixMask1, stat = allocErr)
                      call CheckDeallocation(allocErr, "tempPixMask1")

                      deallocate(tempPixMask2, stat = allocErr)
                      call CheckDeallocation(allocErr, "tempPixMask2")
                
                   end do
                end do

                ! Then exclude everything else in a mod-8 tile with a fluctuation
                !write(*,*) "L1099 mod8xmin=", mod8xmin, " mod8xmax=", mod8xmax, &
                !     " mod8TileDX=",mod8TileDX
                !write(*,*) "L1101 mod8ymin=", mod8ymin, " mod8ymax=", mod8ymax, &
                !     " mod8TileDY=",mod8TileDY
                do x = mod8xmin, mod8xmax, mod8TileDX
                   do y = mod8ymin, mod8ymax, mod8TileDY
                      !if (x + mod8TileDX - 1 .le. xMax &
                      !     .and. y + mod8TileDY - 1 .le. yMax) then 
                      !   pixMask(x : x + mod8TileDX - 1, y : y + mod8TileDY - 1) = &
                      !        all(pixMask(x : x + mod8TileDX - 1, y : y + mod8TileDY - 1))
                      !endif

                      do i1=1,mod8TileDX
                         do j1=1,mod8TileDY
                            if (x+i1-1 .le. xMax  .and. y+j1-1 .le. yMax) then
                               auxlogical=.true.
                               do i2=1,mod8TileDX
                                  do j2=1,mod8TileDY
                                     if (x+i2-1 .le. xMax  .and. y+j2-1 .le. yMax) then
                                        if (.not. pixMask(x+i2-1, y+j2-1)) auxlogical=.false.
                                     endif
                                  enddo
                               enddo
                               pixMask(x+i1-1, y+j1-1) = auxlogical
                            endif
                         enddo
                      enddo
                   end do
                end do

                pixMap = pixMask
     
                ! Then deallocate the pixMask
                deallocate(pixMask, stat = allocErr)
                call CheckDeallocation(allocErr, "pixMask")
   
                ! Allocate space for the modulo-8 tile.
                allocate(mod8Tile(1 : mod8TileDX, 1 : mod8TileDY), stat = allocErr)
                call CheckAllocation(allocErr, "mod8Tile")

                ! Allocate space for the modulo-8 box.
                allocate(meanTile(1 : mod8TileDX, 1 : mod8TileDY), stat = allocErr)
                call CheckAllocation(allocErr, "meanTile")

                !********************************* 
                ! Calculate the modulo-8 map.
                !*********************************

                do x = mod8xmin, mod8xmax, mod8TileDX
                   do y = mod8ymin, mod8ymax, mod8TileDY

                      ! computing the mask only for those part of the image
                      ! where the quality array is Ok, otherwise, skip
                      ! (to prevent cycling in cases of images with empty strips)
                      if (tempPixMask3(x,y)) then
                         do i = 1, mod8TileDX
                            do j = 1, mod8TileDY
                               numPixels = 0
                               n = 0
                               !************************************
                               if (usingFlat) then
                                  do while ((numPixels == 0) .and. (((boxDX + (n * mod8TileDX)) &
                                       * (boxDY + (n * mod8TileDY))) <= size(partialFlat)))
                                     
                                     minDX = max(mod8xmin + i - 1, x + i - 1 - boxDX - (n * mod8TileDX))
                                     minDY = max(mod8ymin + j - 1, y + j - 1 - boxDY - (n * mod8TileDY))
                                     maxDX = min(mod8xmax, x + i - 1 + boxDX + (n * mod8TileDX))
                                     maxDY = min(mod8ymax, y + j - 1 + boxDY + (n * mod8TileDY))
                        
                                     ! Allocate space for pixel mask.
                                     allocate (pixMask(minDX : maxDX, minDY : maxDY), stat = allocErr)
                                     !write(*,*) "L1171 pixMask: minDX=", minDX, " maxDX=", maxDX, &
                                     !     " minDY=", minDY, " maxDY=", maxDY
                                     call CheckAllocation(allocErr, "pixMask")

                                     ! Set pixMask for box
                                     !*********************************************************
                                     ! 2011-05-21 Protecting the bounds of pixMap from crashing
                                     !*********************************************************
                                     if (minDX .lt. xMin) minDX=xMin
                                     if (minDY .lt. yMin) minDY=yMin
                                     if (maxDX .gt. xMax) maxDX=xMax
                                     if (maxDY .gt. yMax) maxDY=yMax
                                     !write(*,*) "L1183 pixMap: minDX=", minDX, " maxDX=", maxDX, &
                                     !     " minDY=", minDY, " maxDY=", maxDY
                                     pixMask = pixMap(minDX : maxDX, minDY : maxDY)
 
 
                              
                                     ! Calculate the number of pixels available for modulo 8 map extraction.
                                     numPixels = count(pixMask(minDX : maxDX : mod8TileDX, &
                                          minDY : maxDY : mod8TileDY))
                                     
                                     n = n + 1

                                     ! If extraction area contains no useable pixels then deallocate array
                                     if (numPixels .eq. 0) then
                                        deallocate(pixMask, stat=allocErr)
                                        call CheckDeallocation(allocErr, "pixMask")
                                     end if
                                  end do
                               else
                                  do while ((numPixels == 0) .and. (((boxDX + (n * mod8TileDX)) &
                                       * (boxDY + (n * mod8TileDY))) <= size(inOSW)))
                                     minDX = max(mod8xmin + i - 1, x + i - 1 - boxDX - (n * mod8TileDX))
                                     minDY = max(mod8ymin + j - 1, y + j - 1 - boxDY - (n * mod8TileDY))
                                     maxDX = min(mod8xmax, x + i - 1 + boxDX + (n * mod8TileDX))
                                     maxDY = min(mod8ymax, y + j - 1 + boxDY + (n * mod8TileDY))
                        
                                     ! Allocate space for pixel mask.
                                     allocate (pixMask(minDX : maxDX, minDY : maxDY), stat = allocErr)
                                     call CheckAllocation(allocErr, "pixMask")
                                  
                                     !*********************************************************
                                     ! 2011-05-21 Protecting the bounds of pixMap from crashing
                                     !*********************************************************
                                     if (minDX .lt. xMin) minDX=xMin
                                     if (minDY .lt. yMin) minDY=yMin
                                     if (maxDX .gt. xMax) maxDX=xMax
                                     if (maxDY .gt. yMax) maxDY=yMax
                                     !write(*,*) "L1220 pixMap: minDX=", minDX, " maxDX=", maxDX, &
                                     !     " minDY=", minDY, " maxDY=", maxDY
                                     ! Set pixMask for box
                                     pixMask = pixMap(minDX : maxDX, minDY : maxDY)

                                     ! Calculate number of pixels available for modulo 8 map extraction.
                                     numPixels = count(pixMask(minDX : maxDX : mod8TileDX, &
                                          minDY : maxDY : mod8TileDY))
                                     n = n + 1

                                     ! If extraction area contains no useable pixels then deallocate array
                                     if (numPixels .eq. 0) then
                                        deallocate(pixMask, stat=allocErr)
                                        call CheckDeallocation(allocErr, "pixMask")
                                     end if
                                  end do
                               endif
                         
                               if (numPixels .gt. 0) then
                               ! numPixels = max(1, numPixels)   ! In case numPixels == 0.

                                  meanTile(i, j) = sum(inOSW(minDX : maxDX : mod8TileDX, &
                                       minDY : maxDY : mod8TileDY), &
                                       mask = pixMask(minDX : maxDX : mod8TileDX, &
                                       minDY : maxDY : mod8TileDY)) / real(numPixels)

                                  deallocate(pixMask, stat = allocErr)
                                  call CheckDeallocation(allocErr, "pixMask")
                               else
                                  meanTile(i, j) = 0.0
                               end if
                            end do
                         end do
                         mean = sum(meanTile) / size(meanTile)
                         if (mean .eq. 0.0) then                                         ! In case mean == 0.
                            mean = 1.0
                         end if
                         mod8Tile = meanTile / mean
                         do i1=1,mod8TileDX
                            do j1=1,mod8TileDY
                               !mod8Map(x : x + mod8TileDX - 1, y : y + mod8TileDY - 1) = mod8Tile
                               if (x+i1-1 .le. xMax .and. y+j1-1 .le. yMax) then
                                  mod8Map(x+i1-1, y+j1-1) = mod8Tile(i1,j1)
                               endif
                            enddo
                         enddo
                      endif ! tempPixMask3
                   end do
                end do

                ! if there are any gaps in the modulo-8 map, fill them with 1.0. 
                where (mod8Map .eq. 0.0) mod8Map = 1.0

                ! If PPS product required.
                if (PPSProduct) then

                   ! Create the PPS Modulo-8 Map.
                   PPSMod8DataSetH = dataSet(PPSMod8Set, Create)

                   ! Copy some of the input blocks and attributes to the output.
                   if (hasOMIMI) then
                      blockH = block(inDataSetH, "OMIMI1", read)
                      call copyBlock(PPSMod8DataSetH, blockH)
                   else
                      blockH = block(inDataSetH, "PRIMARY", read)
                      call copyBlock(PPSMod8DataSetH, blockH)
                   endif

                   sourceH = attributable(inDataSetH)
                   objectH = attributable(PPSMod8DataSetH)
                   call addAttributes(objectH, sourceH)

                   ! Update and add mod 8 specific attributes.
                   call setAttribute(PPSMod8DataSetH, "CONTENT", "OM OSW MODULO-8 TILE", &
                        "Content of file")
                   call setAttribute(PPSMod8DataSetH, "HDUCLASS", "OGIP", &
                        "Format conforms to OGIP/GSFC conventions")
                   call setAttribute(PPSMod8DataSetH, "HDUCLAS1", "IMAGE", "File contains an image")
                   call setAttribute(PPSMod8DataSetH, "HDUCLAS2", "GENERIC", "Detector map")

                   ! Copy some of the input blocks and attributes to the output.
                
                   if (hasBlock(inDataSetH, "MODES")) then
                      blockH = block(inDataSetH, "MODES", read)
                      call copyBlock(PPSMod8DataSetH, blockH)
                   endif
                   !---
                   ! Get a handle and pointer on the PPS mod 8 array and insert the mod 8 data.
                   if (hasOMIMI) then
                      PPSMod8intArrayH = array(PPSMod8DataSetH, "OMIMI1", modify) ! With int32 data
                      PPSMod8intP => int32Array2Data(PPSMod8intArrayH)
                      PPSMod8intP = int(mod8Map)

                      ! Free up memory and pointers.
                      call release(PPSMod8intArrayH)
                      call release(PPSMod8DataSetH)
                   else
                      PPSMod8ArrayH = array(PPSMod8DataSetH, "PRIMARY", modify) ! With float32 data
                      PPSMod8P => real32Array2Data(PPSMod8ArrayH)
                      PPSMod8P = mod8Map

                      ! Free up memory and pointers.
                      call release(PPSMod8ArrayH)
                      call release(PPSMod8DataSetH)
                   endif
                end if

                ! Set tolerance at machine precision
                tol =  1.e-6

                ! Make sure the image begins and ends at a physical CCD pixel boundary
                !---------------------------------------------------------------------
                !
                ! AMK - 16/08/2000
                !
                ! We must force the code below to work even for windows that are not of size modulo 8 by 
                ! excluding part of the window
                !
                !----------------------------------------------------------------------------- 
                !if (modulo(windowX0,8) == 0 .AND. modulo(windowDX,8) == 0 .AND. modulo(windowY0,8) == 0 .AND. modulo(windowDY,8) == 0)then
                
                ! Allocate space for the starting and intermediate image tiles.
                allocate(StartImTile(1 : mod8TileDX, 1 : mod8TileDY), stat = allocErr)
                call CheckAllocation(allocErr, "StartImTile")

                allocate(InterImTile(1 : mod8TileDX, 1 : mod8TileDY), stat = allocErr)
                call CheckAllocation(allocErr, "InterImTile")

                ! Allocate space for the sums over rows and columns
                allocate(TotalInMod8Col(1 : mod8TileDX), stat = allocErr)
                call CheckAllocation(allocErr, "TotalInMod8")

                allocate(TotalInOSWCol(1 : mod8TileDX), stat = allocErr)
                call CheckAllocation(allocErr, "TotalInOSW")

                allocate(TotalInMod8Row(1 : mod8TileDY), stat = allocErr)
                call CheckAllocation(allocErr, "TotalInMod8Row")

                allocate(TotalInOSWRow(1:mod8TileDY), stat=allocErr)
                call CheckAllocation(allocErr, "TotalInOSWRow")

                do x = mod8xmin, mod8xmax, mod8TileDX
                   do y = mod8ymin, mod8ymax, mod8TileDY
                      do i1=1,mod8TileDX
                         do j1=1,mod8TileDY
                            !StartImTile = inOSW(x : x + mod8TileDX - 1, y : y + mod8TileDY - 1)
                            !mod8Tile = mod8Map(x : x + mod8TileDX - 1, y : y + mod8TileDY - 1)
                            if (x+i1-1 .le. xMax .and. y+j1-1 .le. yMax) then
                               StartImTile(i1,j1) = inOSW(x+i1-1, y+j1-1)
                               mod8Tile(i1,j1) = mod8Map(x+i1-1, y+j1-1)
                            endif
                         enddo
                      enddo
                      ! Now dealing with a single mod8Tile
                      ! First redistribute in y (before x)
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      if (.not. hasEXPOSURE) then
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      do i = 1, mod8TileDX
                         TotalInMod8Col(i) = 0.0
                         TotalInOSWCol(i) = 0.0
                         do j = 1, mod8TileDY 

                            ! Set the intermediate image tile to zeros
                            InterImTile(i,j) = 0.0

                            ! Sum up flux in each of the pixels in Col i for the mod8map and real image
                            TotalInMod8Col(i) = TotalInMod8Col(i) + mod8Tile(i,j)
                            TotalInOSWCol(i) = TotalInOSWCol(i) + StartImTile(i,j)
                         end do

                         ! Now find the boundaries and correct image

                         RunningPixel = 1
                         realRunningPixel = real(RunningPixel)
                         RunningMod8Total = 0.0
                         RunningPosition = 0.0
                         OldRunningPosition = 0.0
 
                         do j = 1, mod8TileDY-1
                            do while (RunningMod8Total < TotalInMod8Col(i)*(real(j)-tol) / realMod8TileDY)
                               ! vny 2012.12.13 protecting mod8Tile from a possible subscript error
                               if (RunningPixel .le. mod8TileDY) then 
                                  if (RunningMod8Total + (realRunningPixel - RunningPosition) * mod8Tile(i, RunningPixel) &
                                       <= TotalInMod8Col(i)*real(j) / realMod8TileDY) then

                                     ! Add in the rest of the counts from RunningPixel and move on to the next pixel
                                     RunningMod8Total = RunningMod8Total + (realRunningPixel - RunningPosition) * & 
                                          mod8Tile(i, RunningPixel)
                                     InterImTile(i,j) = InterImTile(i,j) + (realRunningPixel - RunningPosition) * & 
                                          StartImTile(i,RunningPixel)
                                     RunningPosition = realRunningPixel
                                     RunningPixel = RunningPixel + 1
                                     realRunningPixel = real(RunningPixel) 
                                  else 

                                     ! Just add in the neccessary counts
                                     OldRunningPosition = RunningPosition
                                     RunningPosition = OldRunningPosition + (TotalInMod8Col(i) * real(j) / realMod8TileDY &
                                          - RunningMod8Total) / mod8Tile(i, RunningPixel)
                                     InterImTile(i, j) = InterImTile(i, j) + (RunningPosition - OldRunningPosition) & 
                                          * StartImTile(i, RunningPixel)
                                     RunningMod8Total = TotalInMod8Col(i) * real(j) / realMod8TileDY
                                  end if
                               endif ! RinningPixel .le. mod8TileDY
                            end do
                         end do

                         ! now the last pixel contains whatever is left over
                         InterImTile(i, mod8TileDY) = TotalInOSWCol(i) - sum(InterImTile(i, 1: mod8TileDY - 1))

                      end do
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      endif !!!!!  not hasEXPOSURE
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      ! Now redistribute in x
                      StartImTile = InterImTile  
                      do j = 1, mod8TileDY
                         TotalInOSWRow(j) = 0.0
                         do i = 1, mod8TileDX 
                            InterImTile(i,j) = 0.0
                            TotalInOSWRow(j) = TotalInOSWRow(j) + StartImTile(i,j)
                         end do ! i
                      end do ! j
                      TotalInMod8Cols = 0.0
                      do i = 1, mod8TileDX 
                         TotalInMod8Cols = TotalInMod8Cols + TotalInMod8Col(i)
                      end do ! i
                      RunningPixel = 1
                      realRunningPixel = real(RunningPixel)
                      RunningMod8Total = 0.0
                      RunningPosition = 0.0
                      OldRunningPosition = 0.0

                      do i = 1, mod8TileDX-1
                         do while (RunningMod8Total < TotalInMod8Cols * (real(i) - tol) / realMod8TileDX)
                            !vny 2012-12-13 protecting TotalInMod8Col from possible subscript
                            if (RunningPixel .le. mod8TileDX) then
                               if (RunningMod8Total + (realRunningPixel - RunningPosition) * TotalInMod8Col(RunningPixel) &
                                    <= TotalInMod8Cols * real(i) / realMod8TileDX) then
                                  RunningMod8Total = RunningMod8Total + real(RunningPixel - RunningPosition) * & 
                                       TotalInMod8Col(RunningPixel)
                                  do j = 1, mod8TileDY
                                     InterImTile(i,j) = InterImTile(i, j) + (realRunningPixel - RunningPosition) * & 
                                          StartImTile(RunningPixel, j)
                                  end do ! j
                                  RunningPosition = realRunningPixel
                                  RunningPixel = RunningPixel + 1
                                  realRunningPixel = real(RunningPixel)
                               else 
                                  OldRunningPosition = RunningPosition
                                  RunningPosition = OldRunningPosition + (TotalInMod8Cols * real(i) / realMod8TileDX &
                                       - RunningMod8Total) / TotalInMod8Col(RunningPixel)
                                  do j = 1, mod8TileDY
                                     InterImTile(i, j) = InterImTile(i, j) + (RunningPosition - OldRunningPosition)& 
                                          * StartImTile(RunningPixel, j)
                                  end do ! j
                                  RunningMod8Total = TotalInMod8Cols * real(i) / realMod8TileDX
                               end if
                            endif ! RunningPosition .le. mod8TileDX
                         end do
                      end do
                      do j = 1, mod8TileDY
                         InterImTile(mod8TileDX, j) = TotalInOSWRow(j) - sum(InterImTile(1 : mod8TileDX - 1, j))
                      end do
                      !&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& Excluded for testing 
                      if (.not. multiplyFlat .and. .not. usingFlat) then
                      !   ! Write redistributed image to output file
                      do i1=1,mod8TileDX
                         do j1=1,mod8TileDY                      
                            if (x+i1-1 .le. xMax .and. y+j1-1 .le. yMax) then
                               outOSW(x+i1-1, y+j1-1) = InterImTile(i1,j1)
                            endif
                         enddo
                      enddo
                      endif
                      !&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
                      ! Now repeat whole redistribution but with x first then y           
                      !StartImTile = inOSW(x : x + mod8TileDX - 1, y : y + mod8TileDY - 1)
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      !if (.not. hasEXPOSURE) then 
                      !!!!!!!!!!!!!!!!!!!!!!!!!
                      do i1=1,mod8TileDX
                         do j1=1,mod8TileDY   
                            if (x+i1-1 .le. xMax .and. y+j1-1 .le. yMax) then
                               StartImTile(i1,j1) = inOSW(x+i1-1, y+j1-1)
                            endif
                         enddo
                      enddo
                      do j = 1, mod8TileDY
                         TotalInMod8Row(j) = 0.0
                         TotalInOSWRow(j) = 0.0
                         do i = 1, mod8TileDX 
                            InterImTile(i,j) = 0.0
                            TotalInMod8Row(j) = TotalInMod8Row(j) + mod8Tile(i, j)
                            TotalInOSWRow(j) = TotalInOSWRow(j) + StartImTile(i, j)
                         end do
                         RunningPixel = 1
                         RunningMod8Total = 0.0
                         RunningPosition = 0.0
                         OldRunningPosition = 0.0
                         do i = 1, mod8TileDX-1
                            do while (RunningMod8Total < TotalInMod8Row(j)*(real(i)-tol) / realMod8TileDX)
                               ! vny 2012-12-13 protecting mod8Tile from possible subscript errors
                               if (RunningPixel .le.mod8TileDX) then 
                                  if (RunningMod8Total + (real(RunningPixel) - RunningPosition) * mod8Tile(RunningPixel,j) &
                                       <= TotalInMod8Row(j) * real(i) / realMod8TileDX)then
                                     RunningMod8Total = RunningMod8Total + (real(RunningPixel) - RunningPosition) *& 
                                          mod8Tile(RunningPixel, j)
                                     InterImTile(i, j) = InterImTile(i, j) + (real(RunningPixel) - RunningPosition) *& 
                                          StartImTile(RunningPixel, j)
                                     RunningPosition = real(RunningPixel)
                                     RunningPixel = RunningPixel + 1
                                  else 
                                     OldRunningPosition = RunningPosition
                                     RunningPosition = OldRunningPosition + (TotalInMod8Row(j) * real(i) / realMod8TileDX &
                                          - RunningMod8Total) / mod8Tile(RunningPixel, j)
                                     InterImTile(i, j) = InterImTile(i, j) + (RunningPosition - OldRunningPosition) & 
                                          * StartImTile(RunningPixel, j)
                                     RunningMod8Total = TotalInMod8Row(j) * real(i) / realMod8TileDX
                                  end if
                               endif ! RunningPixel .le. mod8TileDX
                            end do
                         end do
                         InterImTile(mod8TileDX, j) = TotalInOSWRow(j) - sum(InterImTile(1 : mod8TileDX - 1, j))
                      end do
 
                      ! Now redistribute in y
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      !if (.not. hasEXPOSURE) then 
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      StartImTile = InterImTile  
                      do i = 1, mod8TileDX
                         TotalInOSWCol(i) = 0.0
                         do j = 1, mod8TileDY 
                            InterImTile(i, j) = 0.0
                            TotalInOSWCol(i) = TotalInOSWCol(i) + StartImTile(i, j)
                         end do ! j
                      end do ! i
                      TotalInMod8Rows = 0.0
                      do j = 1, mod8TileDY
                         TotalInMod8Rows = TotalInMod8Rows + TotalInMod8Row(j)
                      end do ! j

                      RunningPixel = 1
                      realRunningPixel = real(RunningPixel)
                      RunningMod8Total = 0.0
                      RunningPosition = 0.0
                      OldRunningPosition = 0.0

                      do j = 1, mod8TileDY - 1
                         do while (RunningMod8Total < TotalInMod8Rows * (real(j) - tol) / realMod8TileDY)
                            if (RunningMod8Total + (realRunningPixel - RunningPosition) * TotalInMod8Row(RunningPixel) &
                                 <= TotalInMod8Rows * real(j) / realMod8TileDY) then
                               RunningMod8Total = RunningMod8Total + (realRunningPixel - RunningPosition) * & 
                                    TotalInMod8Row(RunningPixel)
                               do i = 1, mod8TileDX
                                  InterImTile(i, j) = InterImTile(i, j) + (realRunningPixel - RunningPosition) * & 
                                       StartImTile(i, RunningPixel)
                               end do ! i
                               RunningPosition = real(RunningPixel)
                               RunningPixel = RunningPixel + 1
                               realRunningPixel = real(RunningPixel)
                            else 
                               OldRunningPosition = RunningPosition
                               RunningPosition = OldRunningPosition + (TotalInMod8Rows * real(j) / realMod8TileDY &
                                    - RunningMod8Total) / TotalInMod8Row(RunningPixel)
                               do i = 1, mod8TileDX
                                  InterImTile(i, j) = InterImTile(i, j) + (RunningPosition - OldRunningPosition) & 
                                       * StartImTile(i, RunningPixel)
                               end do ! i
                               RunningMod8Total = TotalInMod8Rows * real(j) / realMod8TileDY
                            end if
                         end do
                      end do
                      do i = 1, mod8TileDX
                         InterImTile(i, mod8TileDY) = TotalInOSWCol(i) - sum(InterImTile(i, 1 : mod8TileDY - 1))
                      end do
                      !if (.not. multiplyFlat) then
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      !endif !!!!! not hasEXPOSURE
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      
                      !outOSW(x : x + mod8TileDX - 1, y : y + mod8TileDY - 1) = &
                      !     outOSW(x : x + mod8TileDX - 1, y : y + mod8TileDY - 1) + InterImTile
                      do i1=1,mod8TileDX
                         do j1=1,mod8TileDY
                            if (x+i1-1 .le. xMax .and. y+j1-1 .le. yMax) then
                               outOSW(x+i1-1, y+j1-1) = &
                                    outOSW(x+i1-1, y+j1-1) + InterImTile(i1,j1)
                            endif
                         enddo
                      enddo
                      !endif
                      !if (.not. multiplyFlat) then
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                      !endif !!!!! not hasEXPOSURE
                      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                   end do
                end do

                ! Finally divide output file by 2 so that we have an average
                outOSW = outOSW / 2.0


                ! Free up memory.
                deallocate(TotalInMod8Col, stat = allocErr)
                call CheckDeallocation(allocErr, "TotalInMod8Col")
                
                deallocate(TotalInOSWCol, stat=allocErr)
                call CheckDeallocation(allocErr, "TotalInOSWCol")

                deallocate(TotalInMod8Row, stat=allocErr)
                call CheckDeallocation(allocErr, "TotalInMod8Row")

                deallocate(TotalInOSWRow, stat=allocErr)
                call CheckDeallocation(allocErr, "TotalInOSWRow")
             
                deallocate(InterImTile, stat=allocErr)
                call CheckDeallocation(allocErr, "InterImTile")

                deallocate(StartImTile, stat=allocErr)
                call CheckDeallocation(allocErr, "StartImTile")


                !else
                
                ! Image contains only part of some CCD pixels.

                !  write (errstr, *) "Image contains only part of some CCD pixels."
                !  call warning("notCorrectingMod8Noise")
                !  outOSW = inOSW

                !end if
             else
                ! Mean was too small to make good enough mod8 map
                write (errstr, *) "Too few counts in image to remove mod8 noise."
                call warning("LowMod8Noise")
                !if (.not. multiplyFlat) then 
                outOSW = inOSW
                !endif
             endif
             ! VNY Check the output pixels
             where (outOSW .lt. 1.0E-1) outOSW=0.0
      
             ! deallocate bad pixel map
             deallocate(pixMap, stat=allocErr)
             call CheckDeallocation(allocErr, "pixMap")
          else
             ! No Modulo-8 noise in image.
             write (errstr, *) "No Modulo-8 noise in image."
             call warning("noMod8Noise")
             !if (.not. multiplyFlat) then
             outOSW = inOSW
             !endif
          end if
       end if ! .not. multiplyFlat
    else
       write (errstr, *) "Modulo-8 noise is not corrected, as requested"
       call warning("notCorrectingMod8Noise")
       outOSW = inOSW
    endif ! mod8correction 
   

    ! Create the output OM OSW Image file.
    outDataSetH = dataSet(outSet, Create)
    call copyAllAttributes(attributable(inDataSetH), attributable(outDataSetH))

    outArrayH=addArray(outDataSetH, "PRIMARY", real32, outDim, "Counts", "IMAGE")
    qualityArrayH = addArray(outDataSetH, "QUALITY", integer16, outDim, "flags", "IMAGE")

    if (hasBlock(inDataSetH, "MODES")) then  ! ADDED BY CB FRIDAY DECEMBER 20th
        call copyBlock(outDataSetH, block(inDataSetH, "MODES"))  ! ADDED BY CB
    else
        outModesH=addTable(outDataSetH, "MODES", 1, "Binary")
    end if

    objectH=attributable(outDataSetH)
    sourceH = attributable(inDataSetH)
    call addAttributes(objectH, sourceH)

    if(hasOMIMI) then
       blockH=block(inDataSetH, "OMIMI1", read)
       sourceH= attributable(blockH)
       call addAttributes(objectH,sourceH)

       if (hasAttribute(sourceH, "FILTER")) then
          iomFilter= int16Attribute(sourceH, "FILTER")
       else
          iomFilter=0
       endif
    else
       blockH=block(inDataSetH, "PRIMARY", read)
       sourceH=attributable(blockH)
       call addAttributes(objectH, sourceH)
       if (hasAttribute(sourceH, "FILTER")) then
          iomFilter= int16Attribute(sourceH, "FILTER")
       else
          iomFilter=0
       endif
    endif



    if (iomFilter .eq. 200 .or. iomFilter .eq. 1000) then
       call setAttribute(objectH, "DATAMODE", "IMAGING ", "File contains an OM observation image")
    endif
    if (iomFilter .eq. 200)  call setAttribute(outDataSetH, "FILTER", "GRISM2  ")
    if (iomFilter .eq. 1000) call setAttribute(outDataSetH, "FILTER", "GRISM1  ")

    write (detsec, '(a1,i4,a1,i4,a1,i4,a1,i4,a1)') "[",windowX0, ":", (windowX0 + windowDX - 1), &
         & ",",windowY0, ":", (windowY0 + windowDY - 1), "]" 
    write (detsize,'(a3,i4,a3,i4,a1)') "[1:", INT(2048.0/real(binX)),",1:",INT(2048.0/real(binY)), "]"
    call setAttribute(outDataSetH, "DETSIZE", trim(detsize), "Mosaic Keywords")
    call setAttribute(outDataSetH, "DETSEC", trim(detSec), "Mosaic Keywords")

    if (hasExposure) then
       call setAttribute(outDataSetH, "EXPOSURE",averExposure)
    endif
    charAttrValue = stringAttribute(outDataSetH, "DATAMODE") 
    !*******************
    if (.not. hasAttribute(outDataSetH, "FRAMTIME")) then
       call setAttribute(outDataSetH, "FRAMTIME", 7.16800009136875D+00)
    endif

   if (.not. hasAttribute(outDataSetH, "DEADFRAC")) then
       call setAttribute(outDataSetH, "DEADFRAC", 2.42745541206185D-02)
    endif

!--- VNY 23.02.2005 Changing the value of MJDREF in accordance with SPR-3333 (04.06.2004)
!    if (.not. hasAttribute(outDataSetH, "MJDREF")) then
!       call setAttribute(outDataSetH, "MJDREF", 5.08140007312960D+04)
!    endif
    if (.not. hasAttribute(outDataSetH, "MJDREF")) then
       call setAttribute(outDataSetH, "MJDREF", 5.0814D+04)
    endif

    if (.not. hasAttribute(outDataSetH, "TIMEZERO")) then
       call setAttribute(outDataSetH, "TIMEZERO", 0.0D+00)
    endif

    if (.not. hasAttribute(outDataSetH, "RA_SCX")) then
       call setAttribute(outDataSetH, "RA_SCX", 0.0D+00)
    endif

    if (.not. hasAttribute(outDataSetH, "DEC_SCX")) then
       call setAttribute(outDataSetH, "DEC_SCX", 0.0D+00)
    endif
    if (.not. hasAttribute(outDataSetH, "PA")) then
       call setAttribute(outDataSetH, "PA", 0.0D+00)
    endif



    outOSWP => real32Array2Data(outArrayH)
    outOSWP = outOSW 
    ! Release handles on data.
    call release(outArrayH) 
    outQualityArrayP => int16Array2Data(qualityArrayH)
    outQualityArrayP = outQualArray
    call release(qualityArrayH)

    outMode=1



    if (.not. hasBlock(inDataSetH, "MODES")) then ! For the non-grism data just copy MODE block 
       
        outColumnH =  addColumn(outModesH, "MODE", INTEGER32)
        outColumnP => int32Data(outColumnH)
        outColumnP = outMode
        call release(outColumnH) 
        !---
        outMode=windowX0

        outColumnH =  addColumn(outModesH, "WINDOWX0", INTEGER32)
        outColumnP => int32Data(outColumnH)
        outColumnP = outMode
        call release(outColumnH)
        !---
        outMode=windowY0
        outColumnH =  addColumn(outModesH, "WINDOWY0", INTEGER32)
        outColumnP => int32Data(outColumnH)
        outColumnP = outMode
        call release(outColumnH)
        !---
        outMode=windowDX
        outColumnH =  addColumn(outModesH, "WINDOWDX", INTEGER32)
        outColumnP => int32Data(outColumnH)
        outColumnP = outMode
        call release(outColumnH)
        !---
        outMode=windowDY
        outColumnH =  addColumn(outModesH, "WINDOWDY", INTEGER32)
        outColumnP => int32Data(outColumnH)
        outColumnP = outMode
        call release(outColumnH)
        call release(outModesH)
    end if
   
    call release(outDataSetH) 


    !---Creating output partial flat image file
    ! Create the output OM OSW Image file.
    if (usingFlat .and. produceOutputFlat .and. mod8correction) then
       outFlatSetH = dataSet(outFlatSet, Create)

       ! Copy the input blocks and attributes to the output.
       !*******************************************************
       if (hasOMIMI) then
          blockH = block(inDataSetH, "OMIMI1", read)
          call copyBlock(outFlatSetH, blockH)
          sourceH = attributable(inDataSetH)
          objectH = attributable(outFlatSetH)
          call addAttributes(objectH, sourceH)
       else
          blockH = block(inDataSetH, "PRIMARY", read)
          call copyBlock(outFlatSetH, blockH)
          sourceH = attributable(inDataSetH)
          objectH = attributable(outFlatSetH)
          call addAttributes(objectH, sourceH)
       endif

       if (hasBlock(inDataSetH, "RAW")) then
          blockH = block(inDataSetH, "RAW", read)
          call copyBlock(outFlatSetH, blockH)
       endif

       if (hasBlock(inDataSetH, "MODES")) then
          blockH = block(inDataSetH, "MODES", read)
          call copyBlock(outFlatSetH, blockH)
       endif

       if (hasBlock(inDataSetH, "QUALITY")) then
          blockH = block(inDataSetH, "QUALITY", read)
          call copyBlock(outFlatSetH, blockH)
       endif
       ! Get a handle and pointer on the output observation science window array and insert the OSW data.
       if (hasOMIMI) then
          outFlatArrayH = array(outFlatSetH, "OMIMI1", modify) ! With int32 data
          outFlatIntP => int32Array2Data(outFlatArrayH)
          outFlatIntP = int(partialFlat)
       else
          outFlatArrayH = array(outFlatSetH, "PRIMARY", modify)
          outFlatP => real32Array2Data(outFlatArrayH) 
          outFlatP = partialFlat
       endif
       ! Release handles on data.
       call release(outFlatArrayH)
       call release(outFlatSetH)
    endif
    call release(inDataSetH)

    !/////////////////////////////////////////////////////////////////////
    !
    ! AMK - Tidy Memory
    !
    !/////////////////////////////////////////////////////////////////////
    if(allocated(partialFlat)) then
       deallocate(partialFlat, stat = istat)
       call CheckDeallocation(allocErr, "partialFlat")
    end if
    if(allocated(qualArray)) then
       deallocate(qualArray, stat = istat)
       call CheckDeallocation(allocErr, "qualArray")
    end if
    if(allocated(inOSW)) then
       deallocate(inOSW, stat=istat)
       call CheckDeallocation(allocErr, "inOSW")

    end if
    if(allocated(outOSW)) then
       deallocate(outOSW,stat=istat)
       call CheckDeallocation(allocErr, "outOSW")

    end if

    if(allocated(pixMap)) then
       deallocate(pixMap,stat=istat)
       call CheckDeallocation(allocErr, "pixMap")
    end if

    if(allocated(tempPixMask1)) then
       deallocate(tempPixMask1,stat=istat)
       call CheckDeallocation(allocErr, "tempPixMask1")
    end if

    if(allocated(tempPixMask2)) then
       deallocate(tempPixMask2, stat=istat)
       call CheckDeallocation(allocErr, "tempPixMask2")
    end if

    if(allocated(tempPixMask3)) then
       deallocate(tempPixMask3, stat=istat)
       call CheckDeallocation(allocErr, "tempPixMask3")
    end if

    if(allocated(pixMask)) then
       deallocate(pixMask,stat=istat)
       call CheckDeallocation(allocErr, "pixMask")
    end if
    if(allocated(mod8Map)) then
       deallocate(mod8Map,stat=istat)
       call CheckDeallocation(allocErr, "mod8Map")
    end if

    if(allocated(mod8Tile)) then
       deallocate(mod8Tile,stat=istat)
       call CheckDeallocation(allocErr, "mod8Tile")
    end if

    if(allocated(meanTile)) then
       deallocate(meanTile,stat=istat)
       call CheckDeallocation(allocErr, "meanTile")
    end if
    if(allocated(StartImTile)) then
       deallocate(StartImTile,stat=istat)
       call CheckDeallocation(allocErr, "StartImTile")
    end if
    if(allocated(InterImTile)) then
       deallocate(InterImTile,stat=istat)
       call CheckDeallocation(allocErr, "InterImTile")
    end if
    if(allocated(TotalInMod8Col)) then
       deallocate(TotalInMod8Col,stat=istat)
       call CheckDeallocation(allocErr, "TotalInMod8Col")
    end if
    if(allocated(TotalInOSWCol)) then
       deallocate(TotalInOSWCol,stat=istat)
       call CheckDeallocation(allocErr, "TotalInOSWCol")
    end if
    if(allocated(TotalInMod8Row)) then
       deallocate(TotalInMod8Row,stat=istat)
       call CheckDeallocation(allocErr, "TotalInMod8Row")
    end if

    if(allocated(TotalInOSWRow)) then
       deallocate(TotalInOSWRow,stat=istat)
       call CheckDeallocation(allocErr, "TotalInOSWRow")
    end if

  end subroutine ommodmap


!***************************************************
! CheckAllocation
!***************************************************
subroutine CheckAllocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a32, a20)') "Could not allocate memory for : ", string 
       write (errstr, *) comment
       call fatal("allocErr") 
    end if
!    noAllocations = noAllocations + 1
end subroutine CheckAllocation
!***************************************************
! CheckDeallocation
!***************************************************
subroutine CheckDeallocation(status, string)
    integer, intent(in) ::  status
    character(len=*), intent(in) :: string
    character(len=100) :: comment
   
    if(status .ne. 0) then
       write(comment, '(a31, a20)') "Failed to release memory for : ", trim(string) 
       write (errstr, *) comment
       call warning("allocErr", comment) 
    end if
!    noDeallocations = noDeallocations + 1
end subroutine CheckDeallocation

end module ommodmap_mod










